package studentmanagement.api.rest.controller.exception;

import io.jsonwebtoken.ExpiredJwtException;
import jakarta.validation.ConstraintViolationException;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.servlet.mvc.method.annotation.ResponseEntityExceptionHandler;
import studentmanagement.model.course.exception.StudentNotInCourseException;
import studentmanagement.model.course.exception.TeacherNotInCourseException;
import studentmanagement.service.generic.exception.IDableDoesNotExistException;
import studentmanagement.service.grade.exception.InvalidStudentAuthorityException;
import studentmanagement.service.role.exception.InvalidRoleException;

import java.sql.SQLIntegrityConstraintViolationException;
import java.time.LocalDateTime;

@ControllerAdvice
public class GlobalExceptionHandlerController extends ResponseEntityExceptionHandler {

    @ExceptionHandler({
        IllegalArgumentException.class
    })
    public ResponseEntity<Object> handleIllegalArgumentException(IllegalArgumentException e, WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.UNPROCESSABLE_ENTITY.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.UNPROCESSABLE_ENTITY,
            request);
    }

    @ExceptionHandler({
        IDableDoesNotExistException.class
    })
    public ResponseEntity<Object> handleIDableDoesNotExistException(IDableDoesNotExistException e, WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.NOT_FOUND.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.NOT_FOUND,
            request);
    }

    @ExceptionHandler({
        TeacherNotInCourseException.class,
        StudentNotInCourseException.class
    })
    public ResponseEntity<Object> handleConflicts(Exception e, WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.CONFLICT.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.CONFLICT,
            request);
    }

    @ExceptionHandler({
        InvalidStudentAuthorityException.class
    })
    public ResponseEntity<Object> handleInvalidStudentAuthorityException(InvalidStudentAuthorityException e,
                                                                         WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.FORBIDDEN.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.FORBIDDEN,
            request);
    }

    @ExceptionHandler({
        ConstraintViolationException.class
    })
    public ResponseEntity<Object> handleConstraintViolationException(ConstraintViolationException e,
                                                                     WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.BAD_REQUEST.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.BAD_REQUEST,
            request);
    }

    @ExceptionHandler({
        HttpClientErrorException.class
    })
    public ResponseEntity<Object> handleHttpClientErrorException(HttpClientErrorException e,
                                                                 WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            e.getStatusCode().value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), e.getStatusCode(),
            request);
    }

    @ExceptionHandler({
        InvalidRoleException.class
    })
    public ResponseEntity<Object> handleInvalidRoleException(InvalidRoleException e,
                                                             WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.BAD_REQUEST.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.BAD_REQUEST,
            request);
    }

    @ExceptionHandler({
        BadCredentialsException.class
    })
    public ResponseEntity<Object> handleBadCredentialsException(BadCredentialsException e,
                                                                WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            "Invalid username or password.",
            HttpStatus.BAD_REQUEST.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.BAD_REQUEST,
            request);
    }

    @ExceptionHandler({
        ExpiredJwtException.class
    })
    public ResponseEntity<Object> handleExpiredJwtException(ExpiredJwtException e,
                                                            WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            "Your JwtToken has expired. Please Log in again.",
            HttpStatus.BAD_REQUEST.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.BAD_REQUEST,
            request);
    }

    @ExceptionHandler({
        SQLIntegrityConstraintViolationException.class
    })
    public ResponseEntity<Object> handleSQLIntegrityConstraintViolationException(
        SQLIntegrityConstraintViolationException e,
        WebRequest request) {
        ExceptionInformation exceptionInformation = new ExceptionInformation(
            e.getMessage(),
            HttpStatus.CONFLICT.value(),
            LocalDateTime.now(),
            request.getDescription(false)
        );

        return this.handleExceptionInternal(e, exceptionInformation, new HttpHeaders(), HttpStatus.CONFLICT,
            request);
    }
}