package studentmanagement.api.rest.controller.grade;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import studentmanagement.model.API.API;
import studentmanagement.model.course.dto.CourseStudentListDto;
import studentmanagement.model.dto.GradeEntry;
import studentmanagement.model.dto.create.AddGradeRequest;
import studentmanagement.model.dto.find.GradeDto;
import studentmanagement.model.key.IDableKey;
import studentmanagement.service.grade.GradeService;

import java.util.List;
import java.util.function.Supplier;

@RestController
@RequestMapping(API.V1 + API.GRADE)
public class DefaultGradeController {
    private static final Double NO_GRADES = -1.0d;
    private static final String GRADE = "/grade";
    private static final String AVG = "/avg";
    private static final String STUDENTS = "/students";
    private static final String STUDENT = "/student";
    private static final String SORT = "/sort";

    private final GradeService gradeService;


    @Autowired
    public DefaultGradeController(GradeService gradeService) {
        this.gradeService = gradeService;
    }

    @GetMapping(params = "student_id")
    @PreAuthorize(API.ANY_AUTHORITY)
    public ResponseEntity<List<GradeDto>> findByStudentId(@Validated @RequestParam("student_id") IDableKey studentKey) {
        return supplyAndHandleListResponse(() -> gradeService.findByStudentId(studentKey));
    }

    @GetMapping(params = "course_id")
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeDto>> findByCourseId(@Validated @RequestParam("course_id") IDableKey courseKey) {
        return supplyAndHandleListResponse(() -> gradeService.findByCourseId(courseKey));
    }

    @GetMapping(params = {"student_id", "course_id"})
    @PreAuthorize(API.ANY_AUTHORITY)
    public ResponseEntity<List<GradeDto>> findByStudentIdAndCourseId(
        @Validated @RequestParam("student_id") IDableKey studentKey,
        @Validated @RequestParam("course_id") IDableKey courseKey
    ) {
        return supplyAndHandleListResponse(() -> gradeService.findByStudentIdAndCourseId(studentKey, courseKey));
    }

    @GetMapping
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeDto>> findAll() {
        return supplyAndHandleListResponse(gradeService::findAllMappedToDto);
    }

    @DeleteMapping(params = "student_id")
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeDto>> deleteByStudentId(
        @Validated @RequestParam("student_id") IDableKey studentKey) {
        return supplyAndHandleListResponse(() -> gradeService.deleteByStudentId(studentKey));
    }

    @DeleteMapping(params = "course_id")
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeDto>> deleteByCourseId(@Validated @RequestParam("course_id") IDableKey courseKey) {
        return supplyAndHandleListResponse(() -> gradeService.deleteByCourseId(courseKey));
    }

    @DeleteMapping(params = {"student_id", "course_id"})
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeDto>> deleteByStudentIdAndCourseId(
        @Validated @RequestParam("student_id") IDableKey studentKey,
        @Validated @RequestParam("course_id") IDableKey courseKey
    ) {
        return supplyAndHandleListResponse(() -> gradeService.deleteByStudentIdAndCourseId(studentKey, courseKey));
    }

    @PostMapping
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<GradeDto> addGrade(@Validated @RequestBody AddGradeRequest request) {
        return ResponseEntity.ok(gradeService.addGrade(request));
    }

    @GetMapping(value = AVG, params = "student_id")
    @PreAuthorize(API.ANY_AUTHORITY)
    public ResponseEntity<Double> findAverageForStudent(@Validated @RequestParam("student_id") IDableKey studentKey) {
        return supplyAndHandleGradeResponse(() -> gradeService.findAverageForStudent(studentKey));

    }

    @GetMapping(value = AVG, params = "course_id")
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<Double> findAverageForCourse(@Validated @RequestParam("course_id") IDableKey courseKey) {
        return supplyAndHandleGradeResponse(() -> gradeService.findAverageForCourse(courseKey));
    }

    @GetMapping(value = AVG, params = {"student_id", "course_id"})
    @PreAuthorize(API.ANY_AUTHORITY)
    public ResponseEntity<Double> findAverageForStudentInCourse(
        @Validated @RequestParam("student_id") IDableKey studentKey,
        @Validated @RequestParam("course_id") IDableKey courseKey
    ) {
        return supplyAndHandleGradeResponse(() -> gradeService.findAverageForStudentInCourse(studentKey, courseKey));
    }

    @GetMapping(AVG)
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeEntry>> getAverageGradesByStudent() {
        return supplyAndHandleListResponse(gradeService::getAverageGradesByStudent);
    }

    @GetMapping(value = AVG + STUDENTS, params = "course_id")
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<GradeEntry>> getAverageGradesByStudentForCourse(
        @Validated @RequestParam("course_id") IDableKey courseKey) {
        return supplyAndHandleListResponse(() -> gradeService.getAverageGradesByStudentForCourse(courseKey));
    }

    @GetMapping(AVG + STUDENTS + SORT)
    @PreAuthorize(API.ADMIN_TEACHER_AUTHORITY)
    public ResponseEntity<List<CourseStudentListDto>> getStudentsByCourseSortedByGrade() {
        return supplyAndHandleListResponse(gradeService::getStudentsByCourseSortedByGrade);
    }

    private <T> ResponseEntity<List<T>> supplyAndHandleListResponse(Supplier<List<T>> supplier) {
        List<T> list = supplier.get();
        return new ResponseEntity<>(list, (list.isEmpty() ? HttpStatus.NO_CONTENT : HttpStatus.OK));
    }

    private ResponseEntity<Double> supplyAndHandleGradeResponse(Supplier<Double> supplier) {
        Double grade = supplier.get();
        return new ResponseEntity<>(grade, (grade.equals(NO_GRADES) ? HttpStatus.NO_CONTENT : HttpStatus.OK));
    }
}
