package studentmanagement.api.security;

import jakarta.servlet.FilterChain;
import jakarta.servlet.ServletException;
import jakarta.servlet.http.Cookie;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.RequiredArgsConstructor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.lang.NonNull;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.web.authentication.WebAuthenticationDetailsSource;
import org.springframework.stereotype.Component;
import org.springframework.web.filter.OncePerRequestFilter;
import studentmanagement.model.person.Person;
import studentmanagement.model.token.Token;
import studentmanagement.service.cookie.CookieUtil;
import studentmanagement.service.jwt.JwtService;
import studentmanagement.service.user_details.DefaultUserDetailsService;

import java.io.IOException;
import java.util.Optional;

@Component
@RequiredArgsConstructor
public class JwtAuthenticationFilter extends OncePerRequestFilter {
    @Autowired
    private final DefaultUserDetailsService userDetailsService;

    @Autowired
    private final JwtService jwtService;
    @Autowired
    private final CookieUtil cookieUtil;


    @Override
    protected void doFilterInternal(
        @NonNull HttpServletRequest request,
        @NonNull HttpServletResponse response,
        @NonNull FilterChain filterChain
    ) throws ServletException, IOException {
        response.addHeader("Alabala", "Alabala");

        Optional<Cookie> cookieOptional = cookieUtil.findServletCookie(request, CookieUtil.COOKIE_NAME);

        if (cookieOptional.isEmpty()) {
            filterChain.doFilter(request, response);
            return;
        }

        Cookie cookie = cookieOptional.get();

        String jwtToken = cookie.getValue();

        if (jwtToken.equals(JwtService.INVALID_TOKEN)) {
            filterChain.doFilter(request, response);
            return;
        }

        String username = jwtService.extractUsername(jwtToken);

        if (username != null && SecurityContextHolder.getContext().getAuthentication() == null) {
            Person person = this.userDetailsService.loadUserByUsername(username);
            Token token = jwtService.findTokenByPersonOrElse(person, null);

            // Check if this Token is the same as the one in the DB
            if (token == null || !jwtService.isTokenValid(jwtToken, person)) {
                filterChain.doFilter(request, response);
                return;
            }

            if (jwtService.isTokenExpired(jwtToken)) {
                token = renewToken(token, person);
            }

            String cookieString = cookieUtil.createCookie(token.getToken(), CookieUtil.COOKIE_NAME).toString();
            response.setHeader(HttpHeaders.SET_COOKIE, cookieString);

            UsernamePasswordAuthenticationToken authenticationToken =
                new UsernamePasswordAuthenticationToken(
                    person,
                    null,
                    person.getAuthorities()
                );

            authenticationToken.setDetails(
                new WebAuthenticationDetailsSource()
                    .buildDetails(request)
            );

            SecurityContextHolder.getContext().setAuthentication(authenticationToken);

        }

        filterChain.doFilter(request, response);
    }

    private Token renewToken(Token token, Person person) {
        jwtService.delete(token);
        Token newToken = jwtService.generateToken(person);
        jwtService.upsert(newToken);

        logRenew(newToken);

        return newToken;
    }

    private void logRenew(Token token) {
        System.out.println(System.lineSeparator()
                           + System.lineSeparator()
                           + System.lineSeparator()
                           + "Renewing token for User: " + token.getPerson().getUsername()
                           + System.lineSeparator()
                           + System.lineSeparator()
                           + System.lineSeparator()
        );
    }
}
