package studentmanagement.api.security;

import lombok.RequiredArgsConstructor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.HttpMethod;
import org.springframework.security.authentication.AuthenticationProvider;
import org.springframework.security.config.annotation.method.configuration.EnableMethodSecurity;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configurers.AbstractHttpConfigurer;
import org.springframework.security.config.http.SessionCreationPolicy;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.User;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.provisioning.InMemoryUserDetailsManager;
import org.springframework.security.web.SecurityFilterChain;
import org.springframework.security.web.authentication.UsernamePasswordAuthenticationFilter;
import studentmanagement.model.API.API;
import studentmanagement.model.role.Role;

@Configuration
@EnableWebSecurity
@EnableMethodSecurity
@RequiredArgsConstructor
public class SecurityConfig {
    @Autowired
    private final BCryptPasswordEncoder passwordEncoder;

    @Autowired
    private final JwtAuthenticationFilter jwtAuthenticationFilter;

    @Autowired
    private final AuthenticationProvider authenticationProvider;

    @Autowired
    private final LogoutUtil logoutUtil;

    @Bean
    public SecurityFilterChain filterChain(HttpSecurity http) throws Exception {

        http.csrf(AbstractHttpConfigurer::disable)
            .authorizeHttpRequests(config -> config
                .requestMatchers(HttpMethod.POST, API.V1 + API.AUTH + API.ALL).permitAll()
                .requestMatchers(API.V1 + API.ADMIN + API.ALL)
                .hasAnyAuthority(Role.ADMIN.getAuthority(), Role.TEACHER.getAuthority())
                .requestMatchers(HttpMethod.GET)
                .hasAnyAuthority(Role.ADMIN.getAuthority(), Role.TEACHER.getAuthority(), Role.STUDENT.getAuthority())
                .requestMatchers(HttpMethod.PUT).hasAnyAuthority(Role.ADMIN.getAuthority(), Role.TEACHER.getAuthority())
                .requestMatchers(HttpMethod.POST)
                .hasAnyAuthority(Role.ADMIN.getAuthority(), Role.TEACHER.getAuthority())
                .requestMatchers(HttpMethod.DELETE)
                .hasAnyAuthority(Role.ADMIN.getAuthority(), Role.TEACHER.getAuthority())
                .anyRequest().authenticated())
            .addFilterBefore(jwtAuthenticationFilter, UsernamePasswordAuthenticationFilter.class)
            .sessionManagement(sessionManagement ->
                sessionManagement.sessionCreationPolicy(SessionCreationPolicy.STATELESS))
            .authenticationProvider(authenticationProvider)
            .logout(logout -> logout
                .logoutUrl(API.V1 + API.AUTH + API.LOGOUT)
                .addLogoutHandler(logoutUtil)
                .logoutSuccessHandler((
                    (request, response, authentication)
                        -> SecurityContextHolder.clearContext()
                ))
            );

        return http.build();
    }

    @Bean
    public UserDetailsService users() {
        UserDetails admin = User.builder()
            .username("admin")
            .password(passwordEncoder.encode("admin-pass"))
            .roles(Role.ADMIN.getAuthority())
            .build();
        UserDetails teacher = User.builder()
            .username("teacher")
            .password(passwordEncoder.encode("teacher-pass"))
            .roles(Role.TEACHER.getAuthority())
            .build();
        UserDetails student = User.builder()
            .username("student")
            .password(passwordEncoder.encode("student-pass"))
            .roles(Role.STUDENT.getAuthority())
            .build();

        return new InMemoryUserDetailsManager(admin, teacher, student);
    }
}