package studentmanagement.api.rest.course;

import com.google.gson.Gson;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import studentmanagement.model.API.API;
import studentmanagement.api.rest.controller.course.DefaultCourseController;
import studentmanagement.model.course.dto.CourseStudentListDto;
import studentmanagement.model.dto.create.CourseCreateRequest;
import studentmanagement.model.dto.response.CourseCreateResponse;
import studentmanagement.model.dto.find.CourseIdNameDto;
import studentmanagement.model.dto.find.StudentIdNameDto;
import studentmanagement.model.dto.find.TeacherIdNameDto;
import studentmanagement.model.key.IDableKey;
import studentmanagement.service.course.CourseService;

import java.util.List;
import java.util.function.Supplier;

import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@SpringBootApplication
class DefaultCourseControllerTest {
    private static final String COURSE = "/course";

    private CourseService courseService;

    private DefaultCourseController controller;

    @Autowired
    private MockMvc mockMvc;

    private Gson gson = new Gson();

    @BeforeEach
    void setUpTestCase() {
        courseService = Mockito.mock(CourseService.class);
        controller = new DefaultCourseController(courseService);
        mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
    }

    @Test
    void upsert_Returns_Ok() throws Exception {
        TeacherIdNameDto teacher = new TeacherIdNameDto(123L, "Filip");
        CourseCreateRequest request = new CourseCreateRequest("Java", 256L, 123L);
        CourseCreateResponse response = new CourseCreateResponse(1L, "Filip", teacher);

        String requestJson = gson.toJson(request);
        String responseJson = gson.toJson(response);

        Mockito.when(courseService.create(request)).thenReturn(response);

        mockMvc.perform(
                MockMvcRequestBuilders.put(API.V1 + COURSE)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(requestJson)
            ).andExpect(status().isOk())
            .andExpect(content().string(responseJson));
    }

    @Test
    void findById() throws Exception {
        IDableKey request = new IDableKey(1L);
        TeacherIdNameDto teacher = new TeacherIdNameDto(123L, "Filip");
        StudentIdNameDto student1 = new StudentIdNameDto(1L, "Filip1");
        StudentIdNameDto student2 = new StudentIdNameDto(2L, "Filip2");
        StudentIdNameDto student3 = new StudentIdNameDto(3L, "Filip3");
        CourseIdNameDto course = new CourseIdNameDto(1L, "Java");

        CourseStudentListDto response = new CourseStudentListDto(
            course, List.of(student1, student2, student3), teacher
        );
        String responseJson = gson.toJson(response);

        Mockito.when(courseService.findByIdMappedToDto(request)).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + COURSE + "/{id}", 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void getAll_Returns_Ok() throws Exception {
        execute_Mock_And_Test(() -> courseService.findAllMappedToDto(), false);
    }

    @Test
    void getAllSorted_Returns_Ok() throws Exception {
        execute_Mock_And_Test(() -> courseService.findAllSortedByNameMappedToDto(), true);

    }

    private void execute_Mock_And_Test(Supplier<List<CourseStudentListDto>> supplier, boolean sorted) throws Exception {
        TeacherIdNameDto teacher = new TeacherIdNameDto(123L, "Filip");
        StudentIdNameDto student1 = new StudentIdNameDto(1L, "Filip1");
        StudentIdNameDto student2 = new StudentIdNameDto(2L, "Filip2");
        StudentIdNameDto student3 = new StudentIdNameDto(3L, "Filip3");
        CourseIdNameDto course = new CourseIdNameDto(1L, "Java");

        CourseStudentListDto response1 = new CourseStudentListDto(
            course, List.of(student1, student2, student3), teacher
        );

        teacher = new TeacherIdNameDto(456L, "Tosho");
        student1 = new StudentIdNameDto(4L, "Tosho1");
        student2 = new StudentIdNameDto(5L, "Tosho2");
        student3 = new StudentIdNameDto(6L, "Tosho3");
        course = new CourseIdNameDto(2L, "JavaScript");

        CourseStudentListDto response2 = new CourseStudentListDto(
            course, List.of(student1, student2, student3), teacher
        );

        teacher = new TeacherIdNameDto(234L, "Dragan");
        student1 = new StudentIdNameDto(9L, "Dragan1");
        student2 = new StudentIdNameDto(7L, "Dragan2");
        student3 = new StudentIdNameDto(8L, "Dragan3");
        course = new CourseIdNameDto(3L, "C#");

        CourseStudentListDto response3 = new CourseStudentListDto(
            course, List.of(student1, student2, student3), teacher
        );

        List<CourseStudentListDto> response =
            sorted ? List.of(response3, response1, response2)
                : List.of(response1, response2, response3);

        String responseJson = gson.toJson(response);

        Mockito.when(supplier.get()).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + COURSE +
                                       (sorted ? "/sort" : ""), 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }
}