package studentmanagement.api.rest.grade;

import com.google.gson.Gson;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import studentmanagement.model.API.API;
import studentmanagement.api.rest.controller.grade.DefaultGradeController;
import studentmanagement.model.course.dto.CourseStudentListDto;
import studentmanagement.model.dto.GradeEntry;
import studentmanagement.model.dto.create.AddGradeRequest;
import studentmanagement.model.dto.find.CourseIdNameDto;
import studentmanagement.model.dto.find.GradeDto;
import studentmanagement.model.dto.find.StudentIdNameDto;
import studentmanagement.model.dto.find.TeacherIdNameDto;
import studentmanagement.model.key.IDableKey;
import studentmanagement.service.grade.GradeService;

import java.util.List;
import java.util.function.Supplier;

import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

class DefaultGradeControllerTest {

    private static final String GRADE = "/grade";
    private static final String AVG = "/avg";
    private static final String STUDENTS = "/students";
    private static final String STUDENT = "/student";
    private static final String SORT = "/sort";
    private final Gson gson = new Gson();
    private GradeService gradeService;
    private DefaultGradeController controller;
    @Autowired
    private MockMvc mockMvc;

    private TeacherIdNameDto teacher = new TeacherIdNameDto(1L, "Teacher");
    private StudentIdNameDto student1 = new StudentIdNameDto(1L, "Filip");
    private StudentIdNameDto student2 = new StudentIdNameDto(2L, "Filip2");
    private StudentIdNameDto student3 = new StudentIdNameDto(3L, "Filip3");
    private CourseIdNameDto course = new CourseIdNameDto(123L, "Java");
    private GradeDto dto1 = new GradeDto(1L, 5.0, student1, course);
    private GradeDto dto2 = new GradeDto(2L, 4.33, student1, course);
    private GradeDto dto3 = new GradeDto(3L, 3.87, student1, course);
    private List<GradeDto> dtoList = List.of(dto1, dto2, dto3);
    private String responseJson = gson.toJson(dtoList);
    private IDableKey studentKey = new IDableKey(1L);
    private IDableKey courseKey = new IDableKey(123L);

    @BeforeEach
    void setUpTestCase() {
        gradeService = Mockito.mock(GradeService.class);
        controller = new DefaultGradeController(gradeService);
        mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
    }

    @Test
    void findByStudentId_Returns_Ok() throws Exception {
        Mockito.when(gradeService.findByStudentId(studentKey)).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE + "?student_id={id}", 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void findByCourseId_Returns_Ok() throws Exception {
        Mockito.when(gradeService.findByCourseId(courseKey)).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE + "?course_id={id}", 123)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void findByStudentIdAndCourseId_Returns_Ok() throws Exception {
        Mockito.when(gradeService.findByStudentIdAndCourseId(studentKey, courseKey)).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE + "?course_id={student_id}&student_id={course_id}", 123, 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void findAll_Returns_Ok() throws Exception {
        Mockito.when(gradeService.findAllMappedToDto()).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void deleteByStudentId_Returns_Ok() throws Exception {
        Mockito.when(gradeService.deleteByStudentId(studentKey)).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.delete(API.V1 + GRADE + "?student_id={id}", 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void deleteByCourseId_Returns_Ok() throws Exception {
        Mockito.when(gradeService.deleteByCourseId(courseKey)).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.delete(API.V1 + GRADE + "?course_id={id}", 123)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void deleteByStudentIdAndCourseId_Returns_Ok() throws Exception {
        Mockito.when(gradeService.deleteByStudentIdAndCourseId(studentKey, courseKey)).thenReturn(dtoList);

        mockMvc.perform(
            MockMvcRequestBuilders.delete(API.V1 + GRADE + "?course_id={student_id}&student_id={course_id}", 123, 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void addGrade() throws Exception {
        AddGradeRequest request = new AddGradeRequest(
            studentKey, courseKey, dto1.grade()
        );
        String requestJson = gson.toJson(request);
        String responseJson = gson.toJson(dto1);

        Mockito.when(gradeService.addGrade(request)).thenReturn(dto1);

        mockMvc.perform(
            MockMvcRequestBuilders.post(API.V1 + GRADE)
                .contentType(MediaType.APPLICATION_JSON)
                .content(requestJson)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void findAverageForStudent() throws Exception {
        supplyAndValidate(() -> gradeService.findAverageForStudent(studentKey), "?student_id={id}", true);
    }

    @Test
    void findAverageForCourse() throws Exception {
        supplyAndValidate(() -> gradeService.findAverageForCourse(courseKey), "?course_id={id}", false);
    }

    @Test
    void findAverageForStudentInCourse() throws Exception {
        supplyAndValidate(() -> gradeService.findAverageForStudentInCourse(studentKey, courseKey),
            "?student_id={student_id}&course_id={course_id}", true);
    }

    private void supplyAndValidate(Supplier<Double> supplier, String urlParam, boolean isStudent) throws Exception {
        Double grade = 5.43;

        Mockito.when(supplier.get()).thenReturn(grade);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE + AVG + urlParam,
                    isStudent ? 1 : 123, 123)
                .contentType(MediaType.APPLICATION_JSON)
                .content(gson.toJson(grade))
        ).andExpect(status().isOk()).andExpect(content().json("5.43"));
    }

    @Test
    void getAverageGradesByStudent() throws Exception {
        supplyAndValidate(() -> gradeService.getAverageGradesByStudent(), false);
    }

    @Test
    void getAverageGradesByStudentForCourse() throws Exception {
        supplyAndValidate(() -> gradeService.getAverageGradesByStudentForCourse(courseKey), true);
    }

    @Test
    void getStudentsByCourseSortedByGrade() throws Exception {

        List<StudentIdNameDto> students = List.of(student2, student1, student3);
        List<CourseStudentListDto> response = List.of(
            new CourseStudentListDto(course, students, teacher)
        );
        String responseJson = gson.toJson(response);

        Mockito.when(gradeService.getStudentsByCourseSortedByGrade()).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE + AVG + STUDENTS + SORT)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    private void supplyAndValidate(Supplier<List<GradeEntry>> supplier, boolean isCourse) throws Exception {
        GradeEntry entry1 = new GradeEntry(student1, 5.0);
        GradeEntry entry2 = new GradeEntry(student2, 3.33);
        GradeEntry entry3 = new GradeEntry(student3, 6.0);

        List<GradeEntry> response = List.of(entry1, entry2, entry3);
        String responseJson = gson.toJson(response);

        Mockito.when(supplier.get()).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + GRADE + AVG + (isCourse ? (STUDENTS + "?course_id={id}") : ""), 123)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }
}