package studentmanagement.api.rest.student;

import com.google.gson.Gson;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import studentmanagement.model.API.API;
import studentmanagement.api.rest.controller.student.DefaultStudentController;
import studentmanagement.model.dto.create.StudentRegisterRequest;
import studentmanagement.model.dto.response.StudentUpsertResponse;
import studentmanagement.model.dto.find.StudentIdNameDto;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.student.Student;
import studentmanagement.service.student.StudentService;

import java.util.List;

import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

class DefaultStudentControllerTest {
    private static final String STUDENT = "/student";

    private StudentService studentService;

    private DefaultStudentController controller;

    @Autowired
    private MockMvc mockMvc;

    private final Gson gson = new Gson();

    @BeforeEach
    void setUpTestCase() {
        studentService = Mockito.mock(StudentService.class);
        controller = new DefaultStudentController(studentService);
        mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
    }

    @Test
    void findById_Returns_200() throws Exception {
        IDableKey request = new IDableKey(1L);
        StudentIdNameDto response = new StudentIdNameDto(1L, "Filip");

        String responseJson = gson.toJson(response);

        Mockito.when(studentService.findByIdMappedToDto(request)).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + STUDENT + "/{id}", 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void deleteById_Returns_200() throws Exception {
        IDableKey request = new IDableKey(1L);
        Student response = new Student();
        response.setName("Filip");
        response.setId(1L);
        response.setAge(123L);

        String responseJson = gson.toJson(response);

        Mockito.when(studentService.deleteById(request)).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.delete(API.V1 + STUDENT + "/{id}", 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void findAll_Returns_200() throws Exception {
        StudentIdNameDto student1 = new StudentIdNameDto(1L, "Filip");
        StudentIdNameDto student2 = new StudentIdNameDto(2L, "Filip2");
        StudentIdNameDto student3 = new StudentIdNameDto(3L, "Filip3");
        List<StudentIdNameDto> response = List.of(student1, student2, student3);

        String responseJson = gson.toJson(response);

        Mockito.when(studentService.findAllMappedToDto()).thenReturn(response);

        mockMvc.perform(
            MockMvcRequestBuilders.get(API.V1 + STUDENT, 1)
        ).andExpect(status().isOk()).andExpect(content().json(responseJson));
    }

    @Test
    void upsert_Returns_200() throws Exception {
        StudentRegisterRequest request = new StudentRegisterRequest("Filip", 123L);
        StudentUpsertResponse response = new StudentUpsertResponse(1L, "Filip", 123L);

        String requestJson = gson.toJson(request);
        String responseJson = gson.toJson(response);

        Mockito.when(studentService.create(request)).thenReturn(response);

        mockMvc.perform(
                MockMvcRequestBuilders.put(API.V1 + STUDENT)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(requestJson)
            ).andExpect(status().isOk())
            .andExpect(content().string(responseJson));
    }
}