package studentmanagement.service.administrator;

import jakarta.transaction.Transactional;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;
import studentmanagement.data.admin.AdminRepository;
import studentmanagement.mapper.admin.AdminMapper;
import studentmanagement.model.course.Course;
import studentmanagement.model.dto.create.AdminRegisterRequest;
import studentmanagement.model.dto.create.AdminUpsertRequest;
import studentmanagement.model.dto.find.AdminIdNameDto;
import studentmanagement.model.dto.find.StudentCourseIdRequest;
import studentmanagement.model.dto.find.TeacherCourseIdRequest;
import studentmanagement.model.dto.response.AdminUpsertResponse;
import studentmanagement.model.dto.response.LoginResponse;
import studentmanagement.model.dto.response.StudentCourseResponse;
import studentmanagement.model.dto.response.TeacherCourseResponse;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.Person;
import studentmanagement.model.person.admin.Admin;
import studentmanagement.model.person.student.Student;
import studentmanagement.model.person.teacher.Teacher;
import studentmanagement.model.role.Role;
import studentmanagement.model.token.Token;
import studentmanagement.service.course.CourseService;
import studentmanagement.service.generic.DefaultService;
import studentmanagement.service.generic.exception.IDableDoesNotExistException;
import studentmanagement.service.jwt.JwtService;
import studentmanagement.service.student.StudentService;
import studentmanagement.service.teacher.TeacherService;

import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Supplier;

@Service
public class DefaultAdministratorService extends DefaultService<Admin> implements AdministratorService {

    private final StudentService studentService;
    private final AdminRepository adminRepository;
    private final CourseService courseService;
    private final TeacherService teacherService;
    private final AdminMapper adminMapper;
    private final BCryptPasswordEncoder bCryptPasswordEncoder;
    private final JwtService jwtService;

    @Autowired
    public DefaultAdministratorService(StudentService studentService, AdminRepository adminRepository,
                                       CourseService courseService,
                                       TeacherService teacherService, AdminMapper adminMapper,
                                       BCryptPasswordEncoder bCryptPasswordEncoder, JwtService jwtService) {
        super(adminRepository);
        this.adminRepository = adminRepository;
        this.studentService = studentService;
        this.courseService = courseService;
        this.teacherService = teacherService;
        this.adminMapper = adminMapper;
        this.bCryptPasswordEncoder = bCryptPasswordEncoder;
        this.jwtService = jwtService;
    }

    @Override
    public LoginResponse registerAdmin(AdminRegisterRequest request) {
        return login(() -> create(request));
    }

    private LoginResponse login(Supplier<Person> supplier) {
        Person userDetails = supplier.get();

        Token currentToken = jwtService.findTokenByPersonOrElse(userDetails, null);

        if (currentToken != null) {
            return new LoginResponse(currentToken.getToken());
        }

        Token token = jwtService.generateToken(userDetails);

        jwtService.upsert(token);

        return new LoginResponse(token.getToken());
    }

    @Override
    @Transactional
    public StudentCourseResponse enroll(StudentCourseIdRequest dto) {
        return findByIdExecuteAndSave(dto, (student, course) -> course.enroll(student));
    }

    @Override
    @Transactional
    public StudentCourseResponse dismiss(StudentCourseIdRequest dto) {
        return findByIdExecuteAndSave(dto, (student, course) -> course.dismiss(student));
    }

    @Override
    public AdminUpsertResponse upsert(AdminUpsertRequest request) {
        Admin admin = adminUpsertRequestToAdmin(request);

        return adminToAdminCreateResponse(save(admin));
    }

    @Override
    public Admin create(AdminRegisterRequest request) {
        return supplySetAuthorityAndSave(() -> adminRegisterRequestToAdmin(request));

    }

    @Override
    public Admin create(Person person) {
        return supplySetAuthorityAndSave(() -> personToAdmin(person));
    }


    @Override
    public AdminIdNameDto findByIdMappedToDto(IDableKey adminKey) {
        return adminToAdminIdNameDto(findById(adminKey));
    }

    @Override
    public List<AdminIdNameDto> findAllMappedToDto() {
        return allAdminsToAdminIdNameDto(
            adminRepository.findAll()
        );
    }

    @Override
    public Admin findByUsername(String username) {
        return adminRepository.findByUsername(username)
            .orElseThrow(() -> new IDableDoesNotExistException(
                String.format(IDableDoesNotExistException.USERNAME_MESSAGE, username)
            ));
    }

    @Override
    public Admin findByUsernameOrElse(String username, Admin orElse) {
        return adminRepository.findByUsername(username)
            .orElse(orElse);
    }

    @Override
    @Transactional
    public TeacherCourseResponse employ(TeacherCourseIdRequest dto) {
        return findByIdExecuteAndSave(dto, (teacher, course) -> course.employ(teacher));
    }

    @Override
    @Transactional
    public TeacherCourseResponse fire(TeacherCourseIdRequest dto) {
        return findByIdExecuteAndSave(dto, (teacher, course) -> course.fire(teacher));
    }

    private StudentCourseResponse findByIdExecuteAndSave(StudentCourseIdRequest dto,
                                                         BiConsumer<Student, Course> executor) {
        Course course = courseService.findById(dto.courseKey());
        Student student = studentService.findById(dto.studentKey());

        executor.accept(student, course);

        courseService.save(course);

        return adminMapper.newStudentCourseResponse(student, course);
    }

    private TeacherCourseResponse findByIdExecuteAndSave(TeacherCourseIdRequest dto,
                                                         BiConsumer<Teacher, Course> executor) {
        Course course = courseService.findById(dto.courseKey());
        Teacher teacher = teacherService.findById(dto.teacherKey());

        executor.accept(teacher, course);

        courseService.save(course);

        return adminMapper.newTeacherCourseResponse(teacher, course);
    }

    private Admin adminRegisterRequestToAdmin(AdminRegisterRequest request) {
        return adminMapper.adminRegisterRequestToAdmin(request, bCryptPasswordEncoder);
    }

    private Admin adminUpsertRequestToAdmin(AdminUpsertRequest request) {
        return adminMapper.adminUpsertRequestToAdmin(request, bCryptPasswordEncoder);
    }

    private AdminUpsertResponse adminToAdminCreateResponse(Admin admin) {
        return adminMapper.adminToAdminCreateResponse(admin);
    }

    private AdminIdNameDto adminToAdminIdNameDto(Admin admin) {
        return adminMapper.adminToAdminIdNameDto(admin);
    }

    private List<AdminIdNameDto> allAdminsToAdminIdNameDto(List<Admin> admins) {
        return adminMapper.allAdminsToAdminIdNameDto(admins);
    }

    private Admin personToAdmin(Person person) {
        return adminMapper.personToAdmin(person);
    }

    private Admin supplySetAuthorityAndSave(Supplier<Admin> supplier) {
        Admin admin = supplier.get();

        admin.setAuthority(Role.ADMIN);

        return save(admin);
    }
}
