package studentmanagement.service.authentication;

import lombok.RequiredArgsConstructor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.AnonymousAuthenticationToken;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;
import studentmanagement.model.dto.create.AdminLoginRequest;
import studentmanagement.model.dto.create.StudentLoginRequest;
import studentmanagement.model.dto.create.StudentRegisterRequest;
import studentmanagement.model.dto.create.TeacherLoginRequest;
import studentmanagement.model.dto.create.TeacherRegisterRequest;
import studentmanagement.model.dto.response.LoginResponse;
import studentmanagement.model.person.Person;
import studentmanagement.model.token.Token;
import studentmanagement.service.administrator.AdministratorService;
import studentmanagement.service.jwt.JwtService;
import studentmanagement.service.student.StudentService;
import studentmanagement.service.teacher.TeacherService;

import java.util.function.Supplier;

@RequiredArgsConstructor
@Service
public class DefaultAuthenticationService implements AuthenticationService {
    @Autowired
    private final StudentService studentService;
    @Autowired
    private final TeacherService teacherService;
    @Autowired
    private final AdministratorService administratorService;

    @Autowired
    private final AuthenticationManager authenticationManager;

    @Autowired
    private final JwtService jwtService;

    @Override
    public LoginResponse registerStudent(StudentRegisterRequest request) {
        return login(() -> studentService.create(request));
    }

    @Override
    public LoginResponse registerTeacher(TeacherRegisterRequest request) {
        return login(() -> teacherService.create(request));
    }

    private LoginResponse login(Supplier<Person> supplier) {
        Person userDetails = supplier.get();

        Token currentToken = jwtService.findTokenByPersonOrElse(userDetails, null);

        if (currentToken != null) {
            return new LoginResponse(currentToken.getToken());
        }

        Token token = jwtService.generateToken(userDetails);

        jwtService.upsert(token);

        return new LoginResponse(token.getToken());
    }

    @Override
    public LoginResponse loginStudent(StudentLoginRequest request) {
        return authenticate(
            request.getUsername(),
            request.getPassword(),
            () -> studentService.findByUsernameOrElse(request.getUsername(), null)
        );
    }

    @Override
    public LoginResponse loginTeacher(TeacherLoginRequest request) {
        return authenticate(
            request.getUsername(),
            request.getPassword(),
            () -> teacherService.findByUsernameOrElse(request.getUsername(), null)
        );
    }

    @Override
    public LoginResponse loginAdmin(AdminLoginRequest request) {
        return authenticate(
            request.getUsername(),
            request.getPassword(),
            () -> administratorService.findByUsernameOrElse(request.getUsername(), null)
        );
    }

    private LoginResponse authenticate(String username, String password, Supplier<Person> supplier) {
        Authentication authentication = SecurityContextHolder.getContext().getAuthentication();

        if (authentication != null && !(authentication instanceof AnonymousAuthenticationToken)) {
            return new LoginResponse("Already logged in");
        }

        authenticationManager.authenticate(
            new UsernamePasswordAuthenticationToken(
                username,
                password
            )
        );

        return login(supplier);
    }
}
