package studentmanagement.service.grade;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import studentmanagement.data.grade.GradeRepository;
import studentmanagement.model.course.Course;
import studentmanagement.model.course.dto.CourseStudentListDto;
import studentmanagement.model.course.exception.StudentNotInCourseException;
import studentmanagement.model.dto.GradeEntry;
import studentmanagement.model.dto.create.AddGradeRequest;
import studentmanagement.model.dto.find.GradeDto;
import studentmanagement.model.dto.find.StudentIdNameDto;
import studentmanagement.model.grade.Grade;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.student.Student;
import studentmanagement.model.role.Role;
import studentmanagement.service.course.CourseService;
import studentmanagement.service.grade.exception.InvalidStudentAuthorityException;
import studentmanagement.service.jwt.JwtService;
import studentmanagement.service.student.StudentService;
import studentmanagement.service.user_details.DefaultUserDetailsService;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.function.Function;

@Service
public class DefaultGradeService implements GradeService {
    private static final Double NO_GRADES = -1.0d;
    private static final String DOES_NOT_EXIST = "A Grade Table for this Student in this Course does not exist";
    private final GradeRepository gradeRepository;
    private final StudentService studentService;
    private final CourseService courseService;
    private final GradeMapper gradeMapper;
    private final JwtService jwtService;
    private final DefaultUserDetailsService userDetailsService;

    @Autowired
    public DefaultGradeService(GradeRepository gradeRepository, StudentService studentService,
                               CourseService courseService, GradeMapper gradeMapper, JwtService jwtService,
                               DefaultUserDetailsService userDetailsService) {
        this.gradeRepository = gradeRepository;
        this.studentService = studentService;
        this.courseService = courseService;
        this.gradeMapper = gradeMapper;
        this.jwtService = jwtService;
        this.userDetailsService = userDetailsService;
    }

    @Override
    public Grade save(Grade grade) {
        Student student = grade.getStudent();
        Course course = grade.getCourse();

        courseService.save(course);
        studentService.save(student);

        return gradeRepository.save(grade);
    }

    @Override
    public Collection<Grade> saveAll(Collection<Grade> grades) {
        grades.forEach(this::save);

        return grades;
    }

    @Override
    public List<GradeDto> findByStudentId(IDableKey studentKey) {
        validateCurrentUserHasId(studentKey);

        return gradeRepository.findByStudentId(studentKey.id())
            .stream().map(this::gradeToGradeDto).toList();
    }

    @Override
    public List<GradeDto> findByCourseId(IDableKey courseKey) {
        return gradeRepository.findByCourseId(courseKey.id())
            .stream().map(this::gradeToGradeDto).toList();
    }

    @Override
    public List<GradeDto> findByStudentIdAndCourseId(IDableKey studentKey, IDableKey courseKey) {
        validateCurrentUserHasId(studentKey);

        return gradeRepository.findAllByStudentIdAndCourseId(studentKey.id(), courseKey.id())
            .stream().map(this::gradeToGradeDto).toList();
    }

    @Override
    public List<GradeDto> findAllMappedToDto() {
        return gradeRepository.findAll().stream()
            .map(this::gradeToGradeDto).toList();
    }

    @Override
    public List<GradeDto> deleteByStudentId(IDableKey studentKey) {
        return deleteByIdAndSave(gradeRepository.findByStudentId(studentKey.id()));
    }

    @Override
    public List<GradeDto> deleteByCourseId(IDableKey courseKey) {
        return deleteByIdAndSave(gradeRepository.findByCourseId(courseKey.id()));
    }

    @Override
    public List<GradeDto> deleteByStudentIdAndCourseId(IDableKey studentKey, IDableKey courseKey) {
        return deleteByIdAndSave(
            gradeRepository.findAllByStudentIdAndCourseId(
                studentKey.id(), courseKey.id()
            )
        );
    }

    private List<GradeDto> deleteByIdAndSave(List<Grade> grades) {
        gradeRepository.deleteAll(grades);

        return grades
            .stream().map(this::gradeToGradeDto).toList();
    }

    @Override
    public GradeDto addGrade(AddGradeRequest dto) {
        Course course = courseService.findById(dto.getCourseKey());
        Student student = studentService.findById(dto.getStudentKey());

        if (!course.isEnrolled(student)) {
            throw new StudentNotInCourseException(
                "Student with ID:" + student.getId() + " is not enrolled in this Course");
        }

        Grade toAdd = new Grade();
        toAdd.setStudent(student);
        toAdd.setCourse(course);
        toAdd.setGrade(dto.getGrade());

        save(toAdd);

        return gradeMapper.gradeToGradeDto(toAdd);
    }

    @Override
    public Double findAverageForStudent(IDableKey studentKey) {
        validateCurrentUserHasId(studentKey);

        return gradeRepository.findTotalAverageGradeForStudent(studentService.findById(studentKey)).orElse(NO_GRADES);
    }

    @Override
    public Double findAverageForCourse(IDableKey courseKey) {
        return gradeRepository.findTotalAverageGradeForCourse(courseService.findById(courseKey)).orElse(NO_GRADES);
    }

    @Override
    public Double findAverageForStudentInCourse(IDableKey studentKey, IDableKey courseKey) {
        validateCurrentUserHasId(studentKey);

        return gradeRepository.findAverageGradeForStudentInCourse(studentService.findById(studentKey),
            courseService.findById(courseKey)).orElse(NO_GRADES);
    }

    @Override
    public List<GradeEntry> getAverageGradesByStudent() {
        return studentService.findAllMappedToDto().stream()
            .map(mapToGradeEntryWithAverage()).toList();
    }

    @Override
    public List<GradeEntry> getAverageGradesByStudentForCourse(IDableKey iDableKey) {
        return courseService.findByIdMappedToDto(iDableKey).students().stream()
            .map(s -> mapToGradeEntryWithAverageForCourse(iDableKey).apply(s)).toList();
    }

    private GradeDto gradeToGradeDto(Grade grade) {
        return gradeMapper.gradeToGradeDto(grade);
    }

    @Override
    public List<CourseStudentListDto> getStudentsByCourseSortedByGrade() {
        List<CourseStudentListDto> courses = new ArrayList<>(courseService.findAllSortedByNameMappedToDto());

        courses.forEach(c -> c.students().sort((s1, s2) -> Double.compare(
            averageForStudentInCourse(s1.getId(), c.course().getId()),
            averageForStudentInCourse(s2.getId(), c.course().getId())
        )));

        return courses;
    }

    private Double averageForStudentInCourse(Long studentId, Long courseId) {
        return findAverageForStudentInCourse(
            new IDableKey(studentId),
            new IDableKey(courseId)
        );
    }

    private Function<StudentIdNameDto, GradeEntry> mapToGradeEntryWithAverage() {
        return k -> new GradeEntry(k, findAverageForStudent(new IDableKey(k.getId())));
    }

    private Function<StudentIdNameDto, GradeEntry> mapToGradeEntryWithAverageForCourse(IDableKey courseKey) {
        return s ->
            new GradeEntry(
                s, findAverageForStudentInCourse(
                new IDableKey(s.getId()), courseKey
            ));
    }

    private void validateCurrentUserHasId(IDableKey iDableKey) {
        Student student = studentService.findById(iDableKey);

        validateCurrentUserIs(student);
    }

    /**
     * Asserts that if the current User (The one that sent the HttpRequest) is a Student,
     * he is the {@code student} that is being searched
     *
     * @param student
     */
    private void validateCurrentUserIs(Student student) {
        String jwtToken = jwtService.extractJwtTokenFromCurrentRequest()
            .orElseThrow(() -> new HttpClientErrorException(HttpStatus.BAD_REQUEST,
                "No Jwt Token found in Request. Student verification can not be executed"));

        String username = jwtService.extractUsername(jwtToken);

        UserDetails userDetails = userDetailsService.loadUserByUsername(username);

        if (!userDetails.getAuthorities().contains(Role.STUDENT)) {
            return; // Current User is a Teacher or an Admin
        }

        if (!student.getUsername().equals(userDetails.getUsername())) {
            throw new InvalidStudentAuthorityException(
                "You can not view the Grades of other Students"
            );
        }
    }
}
