package studentmanagement.service.jwt;

import io.jsonwebtoken.Claims;
import io.jsonwebtoken.ExpiredJwtException;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;
import io.jsonwebtoken.io.Decoders;
import io.jsonwebtoken.security.Keys;
import jakarta.servlet.http.Cookie;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.transaction.Transactional;
import lombok.RequiredArgsConstructor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;
import studentmanagement.data.TokenRepository;
import studentmanagement.model.person.Person;
import studentmanagement.model.token.Token;
import studentmanagement.service.cookie.CookieUtil;

import java.security.Key;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;

@Service
@RequiredArgsConstructor
public class JwtService {
    public static final String INVALID_TOKEN = "Gepihtitokena";

    private static final long FIVE_MINUTES = 1000 * 60 * 5;
    private static final long TEN_SECONDS = 1000 * 10;
    private static final String SECRET_KEY = "h1tHrnp8s7WQXsrqRIU8Z5CVa9N/sUR4zE5WMLLD5Kk=";
    private static final String AUTHORIZATION_HEADER = "Authorization";
    private static final String BEARER = "Bearer ";

    @Autowired
    private final TokenRepository tokenRepository;
    @Autowired
    private final CookieUtil cookieUtil;

    public String extractUsername(String token) {
        try {
            return extractClaim(token, Claims::getSubject);
        } catch (ExpiredJwtException e) {
            return e.getClaims().getSubject();
        }
    }

    public Date extractExpiration(String token) {
        try {
            return extractClaim(token, Claims::getExpiration);
        } catch (ExpiredJwtException e) {
            return e.getClaims().getExpiration();
        }
    }

    public <T> T extractClaim(String token, Function<Claims, T> claimsResolver) {
        final Claims claims = extractAllClaims(token);
        return claimsResolver.apply(claims);
    }

    private Claims extractAllClaims(String token) {
        return Jwts.parserBuilder()
            .setSigningKey(getSignInKey())
            .build().parseClaimsJws(token)
            .getBody();
    }

    private Key getSignInKey() {
        byte[] keyBytes = Decoders.BASE64.decode(SECRET_KEY);
        return Keys.hmacShaKeyFor(keyBytes);
    }

    public Boolean isTokenExpired(String token) {
        return extractExpiration(token).before(new Date());
    }

    public Token generateToken(Person userDetails) {
        return generateToken(new HashMap<>(), userDetails);
    }

    private Token generateToken(Map<String, Object> extraClaims, Person userDetails) {
        return new Token(Jwts.builder()
            .setClaims(extraClaims)
            .setSubject(userDetails.getUsername())
            .setIssuedAt(new Date())
            .setExpiration(new Date(System.currentTimeMillis() + FIVE_MINUTES))
            .signWith(getSignInKey(), SignatureAlgorithm.HS256)
            .compact(), userDetails);
    }

    public Boolean isLoggedIn(Person userDetails) {
        return tokenRepository.findByPersonUsername(userDetails.getUsername()).isPresent();
    }

    public Boolean validateEquals(String username, String token) {
        Token repoToken = tokenRepository.findByPersonUsername(username).orElse(null);

        if (repoToken == null) {
            return false;
        }

        return token.equals(repoToken.getToken());
    }

    public Boolean isTokenValid(String token, Person userDetails) {
        return validateEquals(userDetails.getUsername(), token);
    }

    public Optional<String> extractJwtTokenFromCurrentRequest() {
        return extractJwtToken(extractCurrentHttpRequestFromContext());
    }

    public Optional<String> extractJwtToken(HttpServletRequest request) {
        Optional<Cookie> cookieOptional = cookieUtil.findServletCookie(request, CookieUtil.COOKIE_NAME);

        if (cookieOptional.isEmpty()) {
            return Optional.empty();
        }

        Cookie cookie = cookieOptional.get();

        String jwtToken = cookie.getValue();

        return Optional.of(jwtToken);
    }

    public HttpServletRequest extractCurrentHttpRequestFromContext() {
        ServletRequestAttributes requestAttributes = (ServletRequestAttributes)
            RequestContextHolder.getRequestAttributes();

        if (requestAttributes == null) {
            throw new HttpClientErrorException(HttpStatus.INTERNAL_SERVER_ERROR,
                "Error while extracting Http RequestAttributes");
        }

        return requestAttributes.getRequest();
    }

    @Transactional
    public synchronized Token upsert(Token token) {
        return tokenRepository.saveAndFlush(token);
    }

    @Transactional
    public synchronized Collection<Token> upsertAll(Collection<Token> tokens) {
        return tokenRepository.saveAllAndFlush(tokens);
    }

    @Transactional
    public synchronized Token delete(Token token) {
        tokenRepository.delete(token);

        return token;
    }

    @Transactional
    public synchronized Collection<Token> deleteAll(Collection<Token> tokens) {
        tokenRepository.deleteAll(tokens);

        return tokens;
    }

    @Transactional
    public synchronized Token findTokenByPersonOrElse(Person userDetails, Token orElse) {
        return findByUsernameOrElse(userDetails.getUsername(), orElse);
    }

    @Transactional
    public synchronized Token findByUsernameOrElse(String username, Token orElse) {
        return tokenRepository.findByPersonUsername(username).orElse(orElse);
    }

    @Transactional
    public synchronized List<Token> findAll() {
        return tokenRepository.findAll();
    }
}
