package studentmanagement.service.student;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;
import studentmanagement.data.student.StudentRepository;
import studentmanagement.mapper.student.StudentMapper;
import studentmanagement.model.dto.create.StudentRegisterRequest;
import studentmanagement.model.dto.create.StudentUpsertRequest;
import studentmanagement.model.dto.find.StudentIdNameDto;
import studentmanagement.model.dto.response.StudentUpsertResponse;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.Person;
import studentmanagement.model.person.student.Student;
import studentmanagement.model.role.Role;
import studentmanagement.service.generic.DefaultService;
import studentmanagement.service.generic.exception.IDableDoesNotExistException;

import java.util.List;
import java.util.function.Supplier;

@Service
public class DefaultStudentService extends DefaultService<Student> implements StudentService {
    private final StudentMapper studentMapper;
    private final StudentRepository studentRepository;
    private final BCryptPasswordEncoder bCryptPasswordEncoder;

    @Autowired
    public DefaultStudentService(StudentRepository studentRepository, StudentMapper studentMapper,
                                 BCryptPasswordEncoder bCryptPasswordEncoder) {
        super(studentRepository);
        this.studentRepository = studentRepository;
        this.studentMapper = studentMapper;
        this.bCryptPasswordEncoder = bCryptPasswordEncoder;
    }

    @Override
    public Student create(StudentRegisterRequest dto) {
        return supplySetAuthorityAndSave(() -> studentRegisterRequestToStudent(dto));

    }

    @Override
    public Student create(Person person) {
        return supplySetAuthorityAndSave(() -> personToStudent(person));

    }

    @Override
    public Student findByUsername(String username) {
        return studentRepository.findByUsername(username)
            .orElseThrow(() -> new IDableDoesNotExistException(
                String.format(IDableDoesNotExistException.USERNAME_MESSAGE, username)
            ));
    }

    @Override
    public Student findByUsernameOrElse(String username, Student orElse) {
        return studentRepository.findByUsername(username)
            .orElse(orElse);
    }

    @Override
    public StudentUpsertResponse upsert(StudentUpsertRequest dto) {
        Student student = studentUpsertRequestToStudent(dto);

        return studentToStudentCreateResponse(save(student));
    }

    @Override
    public StudentIdNameDto findByIdMappedToDto(IDableKey studentKey) {
        return studentToStudentIdNameDto(findById(studentKey));
    }

    @Override
    public List<StudentIdNameDto> findAllMappedToDto() {
        return findAll().stream()
            .map(this::studentToStudentIdNameDto)
            .toList();
    }

    private Student studentRegisterRequestToStudent(StudentRegisterRequest request) {
        return studentMapper.studentCreateRequestToStudent(request, bCryptPasswordEncoder);
    }

    private Student studentUpsertRequestToStudent(StudentUpsertRequest request) {
        return studentMapper.studentUpsertRequestToStudent(request, bCryptPasswordEncoder);
    }

    private StudentIdNameDto studentToStudentIdNameDto(Student student) {
        return studentMapper.studentToStudentIdNameDto(student);
    }

    private StudentUpsertResponse studentToStudentCreateResponse(Student student) {
        return studentMapper.studentToStudentCreateResponse(student);
    }

    private Student personToStudent(Person person) {
        return studentMapper.personToStudent(person);
    }

    private Student supplySetAuthorityAndSave(Supplier<Student> supplier) {
        Student student = supplier.get();

        student.setAuthority(Role.STUDENT);

        return save(student);
    }
}
