package studentmanagement.service.teacher;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;
import studentmanagement.data.teacher.TeacherRepository;
import studentmanagement.mapper.teacher.TeacherMapper;
import studentmanagement.model.dto.create.TeacherRegisterRequest;
import studentmanagement.model.dto.create.TeacherUpsertRequest;
import studentmanagement.model.dto.find.TeacherIdNameDto;
import studentmanagement.model.dto.response.TeacherUpsertResponse;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.Person;
import studentmanagement.model.person.teacher.Teacher;
import studentmanagement.model.role.Role;
import studentmanagement.service.generic.DefaultService;
import studentmanagement.service.generic.exception.IDableDoesNotExistException;

import java.util.List;
import java.util.function.Supplier;

@Service
public class DefaultTeacherService extends DefaultService<Teacher> implements TeacherService {

    private final TeacherMapper teacherMapper;
    private final TeacherRepository teacherRepository;
    private final BCryptPasswordEncoder bCryptPasswordEncoder;

    @Autowired
    public DefaultTeacherService(TeacherRepository teacherRepository, TeacherMapper teacherMapper,
                                 BCryptPasswordEncoder bCryptPasswordEncoder) {
        super(teacherRepository);
        this.teacherRepository = teacherRepository;
        this.teacherMapper = teacherMapper;
        this.bCryptPasswordEncoder = bCryptPasswordEncoder;
    }

    @Override
    public Teacher create(TeacherRegisterRequest dto) {
        return supplySetAuthorityAndSave(() -> teacherRegisterRequestToTeacher(dto));
    }

    @Override
    public Teacher create(Person person) {
        return supplySetAuthorityAndSave(() -> personToTeacher(person));
    }

    @Override
    public Teacher findByUsername(String username) {
        return teacherRepository.findByUsername(username)
            .orElseThrow(() ->
                new IDableDoesNotExistException(
                    String.format(
                        IDableDoesNotExistException.USERNAME_MESSAGE, username
                    )
                )
            );
    }

    @Override
    public TeacherUpsertResponse upsert(TeacherUpsertRequest dto) {
        Teacher teacher = teacherUpsertRequestToTeacher(dto);

        return teacherToTeacherCreateResponse(save(teacher));
    }

    @Override
    public TeacherIdNameDto findByIdMappedToDto(IDableKey teacherKey) {
        return teacherToTeacherIdNameDto(findById(teacherKey));
    }

    @Override
    public Teacher findByUsernameOrElse(String username, Teacher orElse) {
        return teacherRepository.findByUsername(username)
            .orElse(orElse);
    }

    @Override
    public List<TeacherIdNameDto> findAllMappedToDto() {
        return findAll().stream()
            .map(this::teacherToTeacherIdNameDto)
            .toList();
    }

    private Teacher teacherRegisterRequestToTeacher(TeacherRegisterRequest request) {
        return teacherMapper.teacherRegisterRequestToTeacher(request, bCryptPasswordEncoder);
    }

    private Teacher personToTeacher(Person person) {
        return teacherMapper.personToTeacher(person);
    }

    private Teacher teacherUpsertRequestToTeacher(TeacherUpsertRequest request) {
        return teacherMapper.teacherUpsertRequestToTeacher(request, bCryptPasswordEncoder);
    }

    private TeacherIdNameDto teacherToTeacherIdNameDto(Teacher teacher) {
        return teacherMapper.teacherToTeacherIdNameDto(teacher);
    }

    private TeacherUpsertResponse teacherToTeacherCreateResponse(Teacher teacher) {
        return teacherMapper.teacherToTeacherCreateResponse(teacher);
    }

    private Teacher supplySetAuthorityAndSave(Supplier<Teacher> supplier) {
        Teacher teacher = supplier.get();

        teacher.setAuthority(Role.TEACHER);

        return save(teacher);
    }
}
