package studentmanagement.service.administrator;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import studentmanagement.model.course.Course;
import studentmanagement.model.course.exception.StudentNotInCourseException;
import studentmanagement.model.course.exception.TeacherNotInCourseException;
import studentmanagement.model.dto.find.StudentCourseIdRequest;
import studentmanagement.model.dto.find.TeacherCourseIdRequest;
import studentmanagement.model.dto.response.StudentCourseResponse;
import studentmanagement.model.dto.response.TeacherCourseResponse;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.student.Student;
import studentmanagement.model.person.teacher.Degree;
import studentmanagement.model.person.teacher.Teacher;
import studentmanagement.service.course.CourseService;
import studentmanagement.service.generic.exception.IDableDoesNotExistException;
import studentmanagement.service.student.StudentService;
import studentmanagement.service.teacher.TeacherService;

class DefaultAdministratorServiceTest {
    private final String teacherName = "Filip";
    private final String studentName = "Filip2";
    private final String courseName = "Java";

    private final long teacherId = 0;
    private final IDableKey teacherKey = new IDableKey("0");
    private final long studentId = 1;
    private final IDableKey studentKey = new IDableKey("1");
    private final long courseId = 2;
    private final IDableKey courseKey = new IDableKey("2");
    private final long age = 21;
    private final long totalHours = 21;

    private final Teacher teacher = new Teacher(teacherName, Degree.PHD);
    private final Course course = new Course(courseName, totalHours, teacher);
    private final Student student = new Student(studentName, age);
    private TeacherService teacherServiceMock;
    private AdministratorService administratorService;
    private StudentService studentServiceMock;
    private CourseService courseServiceMock;

    @BeforeEach
    void setUpTestCase() {
        teacher.setId(teacherId);
        student.setId(studentId);
        course.setId(courseId);

        courseServiceMock = Mockito.mock(CourseService.class);
        teacherServiceMock = Mockito.mock(TeacherService.class);
        studentServiceMock = Mockito.mock(StudentService.class);

        administratorService = new DefaultAdministratorService(
            studentServiceMock,
            adminRepository, courseServiceMock,
            teacherServiceMock,
            adminMapper, bCryptPasswordEncoder, jwtService);
    }

    @Test
    void enroll_Enrolls_Student_Correctly() {
        StudentCourseResponse expected = new StudentCourseResponse(student, course);

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(studentServiceMock.findById(studentKey)).thenReturn(student);

        StudentCourseResponse actual = administratorService.enroll(new StudentCourseIdRequest(studentKey, courseKey));

        Assertions.assertEquals(expected, actual, "Grade Table not added correctly");
        Assertions.assertTrue(course.getStudents().contains(student), "Student not added to Course");
        course.getStudents().remove(student);
    }

    @Test
    void enroll_Throws_IDableDoesNotExistException_If_Course_Not_Found()
        throws IDableDoesNotExistException {

        Mockito.when(courseServiceMock.findById(courseKey)).thenThrow(
            new IDableDoesNotExistException("Course not found")
        );
        Mockito.when(studentServiceMock.findById(studentKey)).thenReturn(student);

        Assertions.assertThrows(IDableDoesNotExistException.class,
            () -> administratorService.enroll(new StudentCourseIdRequest(studentKey, courseKey)),
            "IDableDoesNotExistException expected");
    }

    @Test
    void dismiss_Dismisses_Student_Correctly()
        throws IDableDoesNotExistException,
        StudentNotInCourseException {
        StudentCourseResponse expected = new StudentCourseResponse(student, course);

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(studentServiceMock.findById(studentKey)).thenReturn(student);

        administratorService.enroll(new StudentCourseIdRequest(studentKey, courseKey));
        StudentCourseResponse actual = administratorService.dismiss(new StudentCourseIdRequest(studentKey, courseKey));

        Assertions.assertEquals(expected, actual, "Grade Table not removed correctly");
        Assertions.assertFalse(course.getStudents().contains(student), "Student not removed from Course");
        course.getStudents().remove(student);
    }

    @Test
    void dismiss_Throws_StudentNotInCourseException()
        throws IDableDoesNotExistException {

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(studentServiceMock.findById(studentKey)).thenReturn(student);

        Assertions.assertThrows(StudentNotInCourseException.class,
            () -> administratorService.dismiss(new StudentCourseIdRequest(studentKey, courseKey)),
            "StudentNotInCourseException expected");
    }

    @Test
    void employ_Employs_Teacher_Correctly() throws IDableDoesNotExistException, TeacherNotInCourseException {
        TeacherCourseResponse expected = new TeacherCourseResponse(teacher, course);

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(teacherServiceMock.findById(teacherKey)).thenReturn(teacher);

        TeacherCourseResponse actual = administratorService.employ(new TeacherCourseIdRequest(teacherKey, courseKey));

        Assertions.assertEquals(expected, actual, "Teacher not employed correctly");
        Assertions.assertTrue(course.getTeacher().isPresent(), "Teacher not added to the course");
        Assertions.assertEquals(course.getTeacher().get(), actual.teacher(), "Teacher not employed correctly");
        course.fire(teacher);
    }

    @Test
    void fire_Fires_Teacher_Correctly() throws IDableDoesNotExistException, TeacherNotInCourseException {
        TeacherCourseResponse expected = new TeacherCourseResponse(teacher, course);

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(teacherServiceMock.findById(teacherKey)).thenReturn(teacher);

        course.employ(teacher);

        TeacherCourseResponse actual = administratorService.fire(new TeacherCourseIdRequest(teacherKey, courseKey));

        Assertions.assertEquals(expected, actual, "Teacher not fired correctly");
        Assertions.assertTrue(course.getTeacher().isEmpty(), "Teacher not removed from the course");
    }

    @Test
    void fire_Throws_TeacherNotInCourseException() throws IDableDoesNotExistException {
        Teacher newTeacher = new Teacher("Someone", Degree.BSC);

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(teacherServiceMock.findById(teacherKey)).thenReturn(newTeacher);

        Assertions.assertThrows(TeacherNotInCourseException.class,
            () -> administratorService.fire(new TeacherCourseIdRequest(teacherKey, courseKey)),
            "TeacherNotInCourseException expected");
    }
}