package studentmanagement.service.grade;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import studentmanagement.data.grade.GradeRepository;
import studentmanagement.model.course.Course;
import studentmanagement.model.dto.create.AddGradeRequest;
import studentmanagement.model.dto.find.StudentCourseIdRequest;
import studentmanagement.model.dto.find.GradeDto;
import studentmanagement.model.grade.Grade;
import studentmanagement.model.key.IDableKey;
import studentmanagement.model.person.student.Student;
import studentmanagement.model.person.teacher.Degree;
import studentmanagement.model.person.teacher.Teacher;
import studentmanagement.service.course.CourseService;
import studentmanagement.service.generic.exception.IDableDoesNotExistException;
import studentmanagement.service.student.StudentService;

class DefaultGradeServiceTest {
    private static final String GRADE = "5.0";
    private final String teacherName = "Filip";
    private final String studentName = "Filip2";
    private final String courseName = "Java";

    private final long teacherId = 0;
    private final IDableKey teacherKey = new IDableKey("0");
    private final long studentId = 1;
    private final IDableKey studentKey = new IDableKey("1");
    private final long courseId = 2;
    private final IDableKey courseKey = new IDableKey("2");
    private final long age = 21;
    private final long totalHours = 21;

    private final Teacher teacher = new Teacher(teacherName, Degree.PHD);
    private final Course course = new Course(courseName, totalHours, teacher);
    private final Student student = new Student(studentName, age);

    private final StudentCourseIdRequest studentCourseIdRequest = new StudentCourseIdRequest(studentKey, courseKey);
    private final AddGradeRequest addGradeRequest = new AddGradeRequest(studentKey, courseKey, GRADE);

    private GradeRepository gradeRepositoryMock;
    private GradeService gradeService;
    private StudentService studentServiceMock;
    private CourseService courseServiceMock;

    @BeforeEach
    void setUpTestCase() {
        teacher.setId(teacherId);
        course.setId(courseId);
        student.setId(studentId);

        courseServiceMock = Mockito.mock(CourseService.class);
        studentServiceMock = Mockito.mock(StudentService.class);
        gradeRepositoryMock = Mockito.mock(GradeRepository.class);
        gradeService = new DefaultGradeService(gradeRepositoryMock, studentServiceMock, courseServiceMock, gradeMapper,
            jwtService, userDetailsService);

    }

    @Test
    void addGrade_Adds_Grade_Correctly() {
        Grade grade = new Grade();
        grade.setStudent(student);
        grade.setCourse(course);
        grade.setGrade(5.0);

        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(studentServiceMock.findById(studentKey)).thenReturn(student);

        Mockito.when(gradeRepositoryMock.save(grade)).thenReturn(grade);
        course.getStudents().add(student);

        GradeDto actual = gradeService.addGrade(addGradeRequest);

        Assertions.assertEquals(grade.getGrade(), actual.grade(), "Grade was not added successfully");
        Assertions.assertEquals(grade.getStudent().getId(), actual.student().id(),
            "Grade Student Id is not the same");
        Assertions.assertEquals(grade.getCourse().getId(), actual.course().id(),
            "Grade Course Id is not the same");
        course.getStudents().remove(student);
    }

    @Test
    void addGrade_Throws_IDableDoesNotExistException_When_Student_Does_Not_Exist() {
        Mockito.when(courseServiceMock.findById(courseKey)).thenReturn(course);
        Mockito.when(studentServiceMock.findById(studentKey)).thenThrow(
            new IDableDoesNotExistException("Student does not exist"));


        Assertions.assertThrows(IDableDoesNotExistException.class,
            () -> gradeService.addGrade(addGradeRequest),
            "IDableDoesNotExistException expected");
    }
}