package studentmanagement.model.course;

import jakarta.persistence.Column;
import jakarta.persistence.Entity;
import jakarta.persistence.FetchType;
import jakarta.persistence.JoinColumn;
import jakarta.persistence.JoinTable;
import jakarta.persistence.ManyToMany;
import jakarta.persistence.ManyToOne;
import jakarta.persistence.Table;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import org.hibernate.Hibernate;
import studentmanagement.model.DefaultIDable;
import studentmanagement.model.course.exception.StudentNotInCourseException;
import studentmanagement.model.course.exception.TeacherNotInCourseException;
import studentmanagement.model.person.student.Student;
import studentmanagement.model.person.teacher.Teacher;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

@Entity
@Table(name = "Courses")
@Setter
@Getter
@AllArgsConstructor
@NoArgsConstructor
public class Course extends DefaultIDable implements Comparable<Course> {
    @Column
    private Long totalHours;
    @Column
    private String name;

    @ManyToMany(fetch = FetchType.EAGER)
    @JoinTable(name = "courses_students",
        joinColumns = {@JoinColumn(name = "course_id", referencedColumnName = "id")},
        inverseJoinColumns = {@JoinColumn(name = "student_id", referencedColumnName = "id")}
    )
    private Set<Student> students = new HashSet<>();

    @ManyToOne(fetch = FetchType.EAGER)
    private Teacher teacher;

    public Collection<Student> getStudents() {
        return students;
    }

    public void employ(Teacher teacher) {
        this.teacher = teacher;
    }

    public void fire(Teacher teacher) throws TeacherNotInCourseException {
        if (!teacher.equals(this.teacher)) {
            throw new TeacherNotInCourseException(
                "Teacher with ID: " + teacher.getId() + " is not employed in this Course");
        }
        this.teacher = null;
    }

    /**
     * Adds a new Student to the Course
     *
     * @param student - the new Student
     */
    public Student enroll(Student student) {
        students.add(student);

        if (Hibernate.isInitialized(student.getCourses())) {
            student.getCourses().add(this);
        }

        return student;
    }

    public Student dismiss(Student student) {
        validateStudentInCourse(student);

        students.remove(student);

        if (Hibernate.isInitialized(student.getCourses())) {
            student.getCourses().remove(this);
        }

        return student;
    }

    public boolean isEnrolled(Student student) {
        return students.contains(student);
    }

    private void validateStudentInCourse(Student student) throws StudentNotInCourseException {
        if (!isEnrolled(student)) {
            throw new StudentNotInCourseException(
                "Student wth ID: " + student.getId() + " is not enrolled in this Course");
        }
    }

    @Override
    public int compareTo(Course o) {
        return getName().compareTo(o.getName());
    }
}
